﻿Imports System.Collections.ObjectModel
Imports System.Collections.Specialized
Imports System.Drawing.Drawing2D
Imports System.Drawing.Text

<Serializable()> _
Public Class clsDrawingPanel

    Inherits Panel

    Public Event eventObjectListChanged(sender As Object, e As NotifyCollectionChangedEventArgs)
    Public Event eventSelectedObjectChanged(SelectedFE As clsDrawingObject)
    Public Event eventHoveredObjectChanged(HoveredFE As clsDrawingObject)
    Public Event eventHighlightedObjectChanged(HighlightedFE As clsDrawingObject)

    Private g As Graphics
    'Private m_MouseOffset As Size
    Private m_MouseDown As Boolean = False
    Private m_MouseDrag As Boolean = False

    Private bufferImage As Bitmap

    Private m_MousePos As Point = Point.Empty
    Public ReadOnly Property MousePos() As Point
        Get
            Return m_MousePos
        End Get

    End Property

    Private m_HoveredObject As clsDrawingObject
    Public Property HoveredObject() As clsDrawingObject
        Get
            Return m_HoveredObject
        End Get
        Set(ByVal value As clsDrawingObject)
            'If value Is m_HoveredObject Then Return
            If Not m_HoveredObject Is Nothing And value Is Nothing Then
                m_HoveredObject.isHovered = False
                RealPaint(m_HoveredObject.DrawingBounds)
            End If
            If Not value Is Nothing Then
                value.isHovered = True
                RealPaint(value.DrawingBounds)
            End If
            If Not m_HoveredObject Is value Then RaiseEvent eventHoveredObjectChanged(value)
            m_HoveredObject = value
            'RealPaint()
        End Set
    End Property

    Private m_HighlightedObject As clsDrawingObject
    Public Property HighlightedObject() As clsDrawingObject
        Get
            Return m_HighlightedObject
        End Get
        Set(ByVal value As clsDrawingObject)
            'If value Is m_HighlightedObject Then Return
            If Not m_HighlightedObject Is Nothing And value Is Nothing Then
                m_HighlightedObject.isHighlighted = False
                RealPaint(m_HighlightedObject.DrawingBounds)
            End If
            If Not value Is Nothing Then
                value.isHighlighted = True
                RealPaint(value.DrawingBounds)
            End If
            If Not m_HighlightedObject Is value Then RaiseEvent eventHighlightedObjectChanged(value)
            m_HighlightedObject = value
            'RealPaint()
        End Set
    End Property

    Private m_SelectedObject As clsDrawingObject
    Public Property SelectedObject() As clsDrawingObject
        Get
            Return m_SelectedObject
        End Get
        Set(ByVal value As clsDrawingObject)
            If value Is m_SelectedObject Then Return
            If Not m_SelectedObject Is Nothing And value Is Nothing Then
                m_SelectedObject.isSelected = False
                RealPaint(m_SelectedObject.DrawingBounds)
            End If
            If Not value Is Nothing Then
                value.isSelected = True
                RealPaint(value.DrawingBounds)
            End If
            If Not m_SelectedObject Is value Then RaiseEvent eventSelectedObjectChanged(value)
            m_SelectedObject = value
            'RealPaint()
        End Set
    End Property

    Private m_colObjects As ObservableCollection(Of clsDrawingObject)
    Public Property colObjects() As ObservableCollection(Of clsDrawingObject)
        Get
            If m_colObjects Is Nothing Then
                m_colObjects = New ObservableCollection(Of clsDrawingObject)
            End If
            Return m_colObjects
        End Get
        Set(ByVal value As ObservableCollection(Of clsDrawingObject))
            m_colObjects = value
        End Set
    End Property

    Private m_DragObjects As New ObservableCollection(Of clsDrawingObject)
    Public ReadOnly Property DragObjects() As ObservableCollection(Of clsDrawingObject)
        Get
            Return m_DragObjects
        End Get
    End Property

    Public Sub New()
        'Me.DoubleBuffered = True
        'SetStyle(ControlStyles.AllPaintingInWmPaint Or _
        '         ControlStyles.OptimizedDoubleBuffer Or _
        '         ControlStyles.UserPaint Or _
        '         ControlStyles.ResizeRedraw, True)
        'UpdateStyles()

        Me.BackColor = Color.White
        Me.BorderStyle = System.Windows.Forms.BorderStyle.FixedSingle

        If m_colObjects Is Nothing Then
            m_colObjects = New ObservableCollection(Of clsDrawingObject)
        End If

        'AddHandler m_colObjects.CollectionChanged, AddressOf ObjectListChanged
        'AddHandler colObjects.CollectionChanged, AddressOf ObjectListChanged
    End Sub

    Public Sub ObjectListChanged(sender As Object, e As NotifyCollectionChangedEventArgs)
        'RaiseEvent eventObjectListChanged(sender, e)
    End Sub

    Private Sub clsDrawingPanel_MouseClick(sender As Object, e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseClick
        If e.Button = Windows.Forms.MouseButtons.Left And Not m_MouseDown Then
            SelectedObject = Nothing
            GetSelected(e.Location)
        End If
    End Sub

    Private Sub clsDrawingPanel_MouseDown(sender As Object, e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        m_MousePos = e.Location
        If e.Button = Windows.Forms.MouseButtons.Left And Not m_MouseDown Then
            m_MouseDown = True
            SelectedObject = Nothing
            GetSelected(e.Location)
            For Each cDO As clsDrawingObject In m_colObjects
                Debug.WriteLine(cDO.Name & "  " & cDO.isHighlighted.ToString)
            Next
            If Not SelectedObject Is Nothing Then
                SelectedObject.Location = New Point(e.Location.X - SelectedObject.MouseOffset.X, _
                                                      e.Location.Y - SelectedObject.MouseOffset.Y)
                RealPaint()

            End If
        End If
    End Sub

    Private Sub clsDrawingPanel_MouseMove(sender As Object, e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseMove
        m_MousePos = e.Location

        GetHovered(e.Location)

        If m_MouseDown And Not SelectedObject Is Nothing Then
            m_MouseDrag = True ' Math.Abs(SelectedObject.OldLocation.X - e.Location.X) > SelectedObject.Bounds.Width / 2 And Math.Abs(SelectedObject.OldLocation.Y - e.Location.Y) > SelectedObject.Bounds.Height / 2 ' True
            SelectedObject.Location = New Point(e.Location.X, _
                                                e.Location.Y)
            If Not SelectedObject.ConnectedTo Is Nothing Then
                If Not SelectedObject.ConnectedTo.ConnectedTo Is Nothing Then
                    SelectedObject.ConnectedTo.ConnectedTo = Nothing
                End If
            End If
            SelectedObject.ConnectedTo = SelectedObject
        End If

        If Not SelectedObject Is Nothing And m_MouseDown And m_MouseDrag Then
            RealPaint() 'SelectedObject.DrawingBounds)
        End If
    End Sub

    Private Sub clsDrawingPanel_MouseUp(sender As Object, e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseUp
        m_MousePos = e.Location
        If Not SelectedObject Is Nothing And m_MouseDrag Then

            If Not HoveredObject Is Nothing Then
                For Each cDO As clsDrawingObject In m_colObjects
                    If cDO.ConnectedTo Is HoveredObject Or cDO.ConnectedTo Is SelectedObject Then
                        cDO.ConnectedTo = Nothing
                    End If
                Next
                If SelectedObject.OldLocation.X <> HoveredObject.OldLocation.X Then
                    SelectedObject.ConnectedTo = HoveredObject
                    HoveredObject.ConnectedTo = SelectedObject
                End If
            Else
                If Not SelectedObject.ConnectedTo Is Nothing Then
                    If Not SelectedObject.ConnectedTo.ConnectedTo Is Nothing Then
                        SelectedObject.ConnectedTo.ConnectedTo = Nothing
                    End If
                End If
                SelectedObject.ConnectedTo = Nothing
            End If

            SelectedObject.Location = SelectedObject.OldLocation
            SelectedObject.isSelected = False
            SelectedObject.isHovered = False
            SelectedObject.isHighlighted = False
            SelectedObject = Nothing
        Else
            GetSelected(e.Location)
        End If

        m_MouseDown = False
        m_MouseDrag = False

        RealPaint()
    End Sub

    Private Sub clsDrawingPanel_Paint(sender As Object, e As System.Windows.Forms.PaintEventArgs) Handles Me.Paint
        Using g = e.Graphics

            If Not bufferImage Is Nothing Then
                g.DrawImage(bufferImage, 0, 0)
            Else
                RealPaint()
            End If

        End Using
    End Sub

    Public Sub RealPaint(Optional RepaintRegion As Rectangle = Nothing)
        Debug.WriteLine("RealPrint")
        Try
            bufferImage = New Bitmap(Me.Bounds.Width, Me.Bounds.Height)
        Catch ex As Exception

        End Try

        Using g As Graphics = Graphics.FromImage(bufferImage)

            g.Clear(Color.White)

            g.SmoothingMode = SmoothingMode.HighQuality
            g.PixelOffsetMode = PixelOffsetMode.HighQuality
            g.TextRenderingHint = TextRenderingHint.ClearTypeGridFit

            If m_colObjects Is Nothing Then Exit Sub

            For Each cDO As clsDrawingObject In m_colObjects
                cDO.ConnectionPainted = False
            Next

            For Each cDO As clsDrawingObject In m_colObjects
                'If cDO.Bounds.IntersectsWith(e.ClipRectangle) Then Return
                If Not cDO.ConnectedTo Is Nothing Then
                    If Not cDO.ConnectionPainted And Not cDO.ConnectedTo.ConnectionPainted Then
                        cDO.ConnectionPainted = True
                        cDO.ConnectedTo.ConnectionPainted = True
                        Dim SPos As Point = cDO.Mittelpunkt(cDO.OldLocation)
                        Dim EPos As Point = cDO.ConnectedTo.Mittelpunkt(cDO.ConnectedTo.Location)
                        Dim TPos As Point = Point.Empty

                        If EPos.X < SPos.X Then
                            TPos = SPos
                            SPos = EPos
                            EPos = TPos
                        End If

                        If cDO.isHovered Or cDO.isHighlighted Or cDO.ConnectedTo.isHovered Or cDO.ConnectedTo.isHighlighted Then

                            PaintSingleConnection(g, Pens.Red, SPos, EPos)
                            'PaintSingleConnection(e, Pens.Red, cDO.Mittelpunkt(cDO.OldLocation), cDO.ConnectedTo.Mittelpunkt(cDO.ConnectedTo.Location))
                        Else
                            If Not SelectedObject Is Nothing Then
                                If cDO Is SelectedObject Then
                                    PaintSingleConnection(g, Pens.SlateGray, SPos, EPos)
                                Else
                                    PaintSingleConnection(g, Pens.Blue, SPos, EPos)
                                End If
                            Else
                                PaintSingleConnection(g, Pens.Blue, SPos, EPos)
                            End If
                            'PaintSingleConnection(e, Pens.Blue, cDO.Mittelpunkt(cDO.OldLocation), cDO.ConnectedTo.Mittelpunkt(cDO.ConnectedTo.Location))
                        End If
                    End If
                End If
                If Not cDO Is SelectedObject Then cDO.Paint(g)
            Next

            If Not SelectedObject Is Nothing Then
                SelectedObject.Paint(g)
            End If

            g.DrawString(Now.ToShortTimeString & " " & Now.Millisecond.ToString, Me.Font, Brushes.Gray, 2, Me.Bounds.Height - g.MeasureString("W", Me.Font).Height)

        End Using

        Me.Invalidate(RepaintRegion)
    End Sub

    Private Sub GetHovered(MPos As Point)
        Dim tmpCDO As clsDrawingObject = Nothing

        If SelectedObject Is Nothing Then

            For Each cDO As clsDrawingObject In m_colObjects
                If Not cDO Is HoveredObject Then cDO.isHovered = False
                cDO.isHighlighted = False
                If cDO.Bounds.Contains(MPos) Then
                    Debug.WriteLine("Hov1")
                    tmpCDO = cDO
                End If
            Next

        Else

            If m_MouseDrag Then

                For Each cDO As clsDrawingObject In m_colObjects
                    If Not cDO Is SelectedObject Then
                        If Not cDO Is HoveredObject Then cDO.isHovered = False
                        cDO.isHighlighted = False
                        If cDO.Bounds.IntersectsWith(SelectedObject.Bounds) Or cDO.Bounds.Contains(MPos) Then
                            Debug.WriteLine("Hov2")
                            tmpCDO = cDO
                        End If
                    End If
                Next

            Else

                For Each cDO As clsDrawingObject In m_colObjects
                    If Not cDO Is SelectedObject Then
                        If Not cDO Is HoveredObject Then cDO.isHovered = False
                        cDO.isHighlighted = False
                        If Not SelectedObject.ConnectedTo Is Nothing Then
                            If cDO Is SelectedObject.ConnectedTo Then
                                cDO.isHighlighted = True
                            End If
                        End If
                        If cDO.Bounds.Contains(MPos) Then
                            Debug.WriteLine("Hov3")
                            tmpCDO = cDO
                        End If
                    Else
                        'Stop
                    End If
                Next

            End If

        End If

        HoveredObject = tmpCDO

        If Not HoveredObject Is Nothing Then
            If Not HoveredObject.ConnectedTo Is Nothing Then
                HighlightedObject = HoveredObject.ConnectedTo
            Else
                HighlightedObject = Nothing
            End If
        Else
            HighlightedObject = Nothing
        End If
    End Sub

    Private Sub GetSelected(MPos As Point)
        Dim tmpCDO As clsDrawingObject = Nothing

        For Each cDO As clsDrawingObject In m_colObjects
            cDO.isHovered = False
            cDO.isHighlighted = False
            cDO.isSelected = False
            If cDO.Bounds.Contains(MPos) Then
                tmpCDO = cDO
            End If
        Next

        SelectedObject = tmpCDO

        If Not SelectedObject Is Nothing Then
            If Not SelectedObject.ConnectedTo Is Nothing Then
                HighlightedObject = SelectedObject.ConnectedTo
            Else
                HighlightedObject = Nothing
            End If
        End If
    End Sub

    Public Sub PaintSingleConnection(ByRef g As Graphics, Stift As Pen, ByVal StartPunkt As Point, ByVal EndPunkt As Point)
        Try
            Dim Abstand As Integer = 20
            If Math.Abs(EndPunkt.X - StartPunkt.X) < Abstand Then
                Abstand = Math.Abs(EndPunkt.X - StartPunkt.X)
            End If

            Dim Start As Point = New Point(StartPunkt.X + Abstand, StartPunkt.Y)
            Dim Ende As Point = New Point(EndPunkt.X - Abstand, EndPunkt.Y)

            Dim Mitte As Point

            Mitte = New Point(CInt(StartPunkt.X + (EndPunkt.X - StartPunkt.X) / 2), CInt(StartPunkt.Y + (EndPunkt.Y - StartPunkt.Y) / 2))

            Dim Mitte2 As Point

            Mitte2 = New Point(CInt(StartPunkt.X + (EndPunkt.X - StartPunkt.X) / 4), CInt(StartPunkt.Y + (EndPunkt.Y - StartPunkt.Y) / 2))
            g.DrawBezier(Stift, _
                                  New Point(StartPunkt.X, StartPunkt.Y), _
                                 New Point(Start.X, Start.Y), _
                                 New Point(Mitte2.X, Start.Y), _
                                 Mitte)

            Mitte2 = New Point(CInt(StartPunkt.X + ((EndPunkt.X - StartPunkt.X) / 4) * 3), CInt(StartPunkt.Y + (EndPunkt.Y - StartPunkt.Y) / 2))
            g.DrawBezier(Stift, _
                                  Mitte, _
                                  New Point(Mitte2.X, Ende.Y), _
                                  New Point(Ende.X, Ende.Y), _
                                  New Point(EndPunkt.X, EndPunkt.Y))
        Catch ex As Exception

        End Try
    End Sub

End Class
